import json
from typing import Any

import pytest

from . import expression, value

owner_type_info = {
    "Owner": 'STRUCT("DisplayName" VARCHAR,"ID" VARCHAR)',
    "aws_region": "VARCHAR",
}

bucket_type_info = {
    "Buckets": 'STRUCT("Name" VARCHAR,"CreationDate" TIMESTAMP_S)[]',
    "aws_region": "VARCHAR",
}


@pytest.mark.parametrize(
    "parser_cls,input",
    [
        (
            expression.BoundConstant,
            {
                "expression_class": "BOUND_CONSTANT",
                "type": "VALUE_CONSTANT",
                "alias": "",
                "query_location": 18446744073709552000,
                "value": {
                    "type": {
                        "id": "MAP",
                        "type_info": {
                            "type": "LIST_TYPE_INFO",
                            "alias": "",
                            "extension_info": None,
                            "child_type": {
                                "id": "STRUCT",
                                "type_info": {
                                    "type": "STRUCT_TYPE_INFO",
                                    "alias": "",
                                    "extension_info": None,
                                    "child_types": [
                                        {
                                            "first": "key",
                                            "second": {
                                                "id": "VARCHAR",
                                                "type_info": None,
                                            },
                                        },
                                        {
                                            "first": "value",
                                            "second": {
                                                "id": "VARCHAR",
                                                "type_info": None,
                                            },
                                        },
                                    ],
                                },
                            },
                        },
                    },
                    "is_null": False,
                    "value": {
                        "children": [
                            {
                                "type": {
                                    "id": "STRUCT",
                                    "type_info": {
                                        "type": "STRUCT_TYPE_INFO",
                                        "alias": "",
                                        "extension_info": None,
                                        "child_types": [
                                            {
                                                "first": "key",
                                                "second": {
                                                    "id": "VARCHAR",
                                                    "type_info": None,
                                                },
                                            },
                                            {
                                                "first": "value",
                                                "second": {
                                                    "id": "VARCHAR",
                                                    "type_info": None,
                                                },
                                            },
                                        ],
                                    },
                                },
                                "is_null": False,
                                "value": {
                                    "children": [
                                        {
                                            "type": {
                                                "id": "VARCHAR",
                                                "type_info": None,
                                            },
                                            "is_null": False,
                                            "value": "color",
                                        },
                                        {
                                            "type": {
                                                "id": "VARCHAR",
                                                "type_info": None,
                                            },
                                            "is_null": False,
                                            "value": "red",
                                        },
                                    ]
                                },
                            }
                        ]
                    },
                },
            },
        ),
        (
            expression.BoundFunction,
            {
                "expression_class": "BOUND_FUNCTION",
                "type": "BOUND_FUNCTION",
                "alias": "",
                "query_location": 40,
                "return_type": {
                    "id": "STRUCT",
                    "type_info": {
                        "type": "STRUCT_TYPE_INFO",
                        "alias": "",
                        "extension_info": None,
                        "child_types": [
                            {
                                "first": "color",
                                "second": {"id": "VARCHAR", "type_info": None},
                            }
                        ],
                    },
                },
                "children": [
                    {
                        "expression_class": "BOUND_FUNCTION",
                        "type": "BOUND_FUNCTION",
                        "alias": "color",
                        "query_location": 49,
                        "return_type": {"id": "VARCHAR", "type_info": None},
                        "children": [
                            {
                                "expression_class": "BOUND_COLUMN_REF",
                                "type": "BOUND_COLUMN_REF",
                                "alias": "mapping",
                                "query_location": 18446744073709552000,
                                "return_type": {
                                    "id": "MAP",
                                    "type_info": {
                                        "type": "LIST_TYPE_INFO",
                                        "alias": "",
                                        "extension_info": None,
                                        "child_type": {
                                            "id": "STRUCT",
                                            "type_info": {
                                                "type": "STRUCT_TYPE_INFO",
                                                "alias": "",
                                                "extension_info": None,
                                                "child_types": [
                                                    {
                                                        "first": "key",
                                                        "second": {
                                                            "id": "VARCHAR",
                                                            "type_info": None,
                                                        },
                                                    },
                                                    {
                                                        "first": "value",
                                                        "second": {
                                                            "id": "VARCHAR",
                                                            "type_info": None,
                                                        },
                                                    },
                                                ],
                                            },
                                        },
                                    },
                                },
                                "binding": {"table_index": 0, "column_index": 0},
                                "depth": 0,
                            },
                            {
                                "expression_class": "BOUND_CONSTANT",
                                "type": "VALUE_CONSTANT",
                                "alias": "",
                                "query_location": 18446744073709552000,
                                "value": {
                                    "type": {"id": "VARCHAR", "type_info": None},
                                    "is_null": False,
                                    "value": "color",
                                },
                            },
                        ],
                        "name": "map_extract_value",
                        "arguments": [
                            {"id": "ANY", "type_info": None},
                            {"id": "VARCHAR", "type_info": None},
                        ],
                        "original_arguments": [],
                        "catalog_name": "system",
                        "schema_name": "main",
                        "has_serialize": False,
                        "is_operator": False,
                    }
                ],
                "name": "struct_pack",
                "arguments": [],
                "original_arguments": [],
                "catalog_name": "system",
                "schema_name": "main",
                "has_serialize": True,
                "function_data": {
                    "variable_return_type": {
                        "id": "STRUCT",
                        "type_info": {
                            "type": "STRUCT_TYPE_INFO",
                            "alias": "",
                            "extension_info": None,
                            "child_types": [
                                {
                                    "first": "color",
                                    "second": {"id": "VARCHAR", "type_info": None},
                                }
                            ],
                        },
                    }
                },
                "is_operator": False,
            },
        ),
        (
            value.ValueType_map,
            {
                "id": "MAP",
                "type_info": {
                    "type": "LIST_TYPE_INFO",
                    "alias": "",
                    "extension_info": None,
                    "child_type": {
                        "id": "STRUCT",
                        "type_info": {
                            "type": "STRUCT_TYPE_INFO",
                            "alias": "",
                            "extension_info": None,
                            "child_types": [
                                {
                                    "first": "key",
                                    "second": {"id": "VARCHAR", "type_info": None},
                                },
                                {
                                    "first": "value",
                                    "second": {"id": "VARCHAR", "type_info": None},
                                },
                            ],
                        },
                    },
                },
            },
        ),
        (
            expression.BoundColumnRef,
            {
                "expression_class": "BOUND_COLUMN_REF",
                "type": "BOUND_COLUMN_REF",
                "alias": "mapping",
                "query_location": 18446744073709551615,
                "return_type": {
                    "id": "MAP",
                    "type_info": {
                        "type": "LIST_TYPE_INFO",
                        "alias": "",
                        "extension_info": None,
                        "child_type": {
                            "id": "STRUCT",
                            "type_info": {
                                "type": "STRUCT_TYPE_INFO",
                                "alias": "",
                                "extension_info": None,
                                "child_types": [
                                    {
                                        "first": "key",
                                        "second": {"id": "VARCHAR", "type_info": None},
                                    },
                                    {
                                        "first": "value",
                                        "second": {"id": "VARCHAR", "type_info": None},
                                    },
                                ],
                            },
                        },
                    },
                },
                "binding": {"table_index": 0, "column_index": 0},
                "depth": 0,
            },
        ),
        (
            expression.BoundFunction,
            {
                "expression_class": "BOUND_FUNCTION",
                "type": "BOUND_FUNCTION",
                "alias": "",
                "query_location": 39,
                "return_type": {"id": "VARCHAR", "type_info": None},
                "children": [
                    {
                        "expression_class": "BOUND_COLUMN_REF",
                        "type": "BOUND_COLUMN_REF",
                        "alias": "mapping",
                        "query_location": 18446744073709551615,
                        "return_type": {
                            "id": "MAP",
                            "type_info": {
                                "type": "LIST_TYPE_INFO",
                                "alias": "",
                                "extension_info": None,
                                "child_type": {
                                    "id": "STRUCT",
                                    "type_info": {
                                        "type": "STRUCT_TYPE_INFO",
                                        "alias": "",
                                        "extension_info": None,
                                        "child_types": [
                                            {
                                                "first": "key",
                                                "second": {
                                                    "id": "VARCHAR",
                                                    "type_info": None,
                                                },
                                            },
                                            {
                                                "first": "value",
                                                "second": {
                                                    "id": "VARCHAR",
                                                    "type_info": None,
                                                },
                                            },
                                        ],
                                    },
                                },
                            },
                        },
                        "binding": {"table_index": 0, "column_index": 0},
                        "depth": 0,
                    },
                    {
                        "expression_class": "BOUND_CONSTANT",
                        "type": "VALUE_CONSTANT",
                        "alias": "",
                        "query_location": 18446744073709551615,
                        "value": {
                            "type": {"id": "VARCHAR", "type_info": None},
                            "is_null": False,
                            "value": "color",
                        },
                    },
                ],
                "name": "map_extract_value",
                "arguments": [
                    {"id": "ANY", "type_info": None},
                    {"id": "VARCHAR", "type_info": None},
                ],
                "original_arguments": [],
                "catalog_name": "system",
                "schema_name": "main",
                "has_serialize": False,
                "is_operator": False,
            },
        ),
        (
            expression.BoundComparison,
            {
                "expression_class": "BOUND_COMPARISON",
                "type": "COMPARE_EQUAL",
                "alias": "",
                "query_location": 18446744073709551615,
                "left": {
                    "expression_class": "BOUND_FUNCTION",
                    "type": "BOUND_FUNCTION",
                    "alias": "",
                    "query_location": 39,
                    "return_type": {"id": "VARCHAR", "type_info": None},
                    "children": [
                        {
                            "expression_class": "BOUND_COLUMN_REF",
                            "type": "BOUND_COLUMN_REF",
                            "alias": "mapping",
                            "query_location": 18446744073709551615,
                            "return_type": {
                                "id": "MAP",
                                "type_info": {
                                    "type": "LIST_TYPE_INFO",
                                    "alias": "",
                                    "extension_info": None,
                                    "child_type": {
                                        "id": "STRUCT",
                                        "type_info": {
                                            "type": "STRUCT_TYPE_INFO",
                                            "alias": "",
                                            "extension_info": None,
                                            "child_types": [
                                                {
                                                    "first": "key",
                                                    "second": {
                                                        "id": "VARCHAR",
                                                        "type_info": None,
                                                    },
                                                },
                                                {
                                                    "first": "value",
                                                    "second": {
                                                        "id": "VARCHAR",
                                                        "type_info": None,
                                                    },
                                                },
                                            ],
                                        },
                                    },
                                },
                            },
                            "binding": {"table_index": 0, "column_index": 0},
                            "depth": 0,
                        },
                        {
                            "expression_class": "BOUND_CONSTANT",
                            "type": "VALUE_CONSTANT",
                            "alias": "",
                            "query_location": 18446744073709551615,
                            "value": {
                                "type": {"id": "VARCHAR", "type_info": None},
                                "is_null": False,
                                "value": "color",
                            },
                        },
                    ],
                    "name": "map_extract_value",
                    "arguments": [
                        {"id": "ANY", "type_info": None},
                        {"id": "VARCHAR", "type_info": None},
                    ],
                    "original_arguments": [],
                    "catalog_name": "system",
                    "schema_name": "main",
                    "has_serialize": False,
                    "is_operator": False,
                },
                "right": {
                    "expression_class": "BOUND_CONSTANT",
                    "type": "VALUE_CONSTANT",
                    "alias": "",
                    "query_location": 18446744073709551615,
                    "value": {
                        "type": {"id": "VARCHAR", "type_info": None},
                        "is_null": False,
                        "value": "red",
                    },
                },
            },
        ),
        (
            value.ValueType_list,
            {
                "id": "LIST",
                "type_info": {
                    "type": "LIST_TYPE_INFO",
                    "alias": "",
                    "extension_info": None,
                    "child_type": {"id": "INTEGER", "type_info": None},
                },
            },
        ),
        (
            expression.BoundFunction,
            {
                "expression_class": "BOUND_FUNCTION",
                "type": "BOUND_FUNCTION",
                "alias": "",
                "query_location": 18446744073709552000,
                "return_type": {"id": "INTEGER", "type_info": None},
                "children": [
                    {
                        "expression_class": "BOUND_COLUMN_REF",
                        "type": "BOUND_COLUMN_REF",
                        "alias": "switches",
                        "query_location": 18446744073709552000,
                        "return_type": {
                            "id": "LIST",
                            "type_info": {
                                "type": "LIST_TYPE_INFO",
                                "alias": "",
                                "extension_info": None,
                                "child_type": {"id": "INTEGER", "type_info": None},
                            },
                        },
                        "binding": {"table_index": 0, "column_index": 0},
                        "depth": 0,
                    },
                    {
                        "expression_class": "BOUND_CONSTANT",
                        "type": "VALUE_CONSTANT",
                        "alias": "",
                        "query_location": 18446744073709552000,
                        "value": {
                            "type": {"id": "BIGINT", "type_info": None},
                            "is_null": False,
                            "value": 0,
                        },
                    },
                ],
                "name": "array_extract",
                "arguments": [
                    {
                        "id": "LIST",
                        "type_info": {
                            "type": "LIST_TYPE_INFO",
                            "alias": "",
                            "extension_info": None,
                            "child_type": {"id": "INTEGER", "type_info": None},
                        },
                    },
                    {"id": "BIGINT", "type_info": None},
                ],
                "original_arguments": [],
                "catalog_name": "system",
                "schema_name": "main",
                "has_serialize": False,
                "is_operator": False,
            },
        ),
    ],
)
def test_parse(parser_cls: Any, input: dict[str, Any]) -> None:
    parser_cls.model_validate(input)


@pytest.mark.parametrize(
    "json_source,expected_sql,expected_bound_types",
    [
        (
            '{"filters":[{"expression_class":"BOUND_FUNCTION","type":"BOUND_FUNCTION","alias":"","query_location":18446744073709551615,"return_type":{"id":"BOOLEAN","type_info":null},"children":[{"expression_class":"BOUND_COLUMN_REF","type":"BOUND_COLUMN_REF","alias":"aws_region","query_location":70,"return_type":{"id":"VARCHAR","type_info":null},"binding":{"table_index":0,"column_index":0},"depth":0},{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"bar"}}],"name":"prefix","arguments":[{"id":"VARCHAR","type_info":null},{"id":"VARCHAR","type_info":null}],"original_arguments":[],"has_serialize":false,"is_operator":false},{"expression_class":"BOUND_COMPARISON","type":"COMPARE_EQUAL","alias":"","query_location":18446744073709551615,"left":{"expression_class":"BOUND_COLUMN_REF","type":"BOUND_COLUMN_REF","alias":"aws_region","query_location":18446744073709551615,"return_type":{"id":"VARCHAR","type_info":null},"binding":{"table_index":0,"column_index":0},"depth":0},"right":{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"us-east-1"}}}],"column_binding_names_by_index":["aws_region","Buckets","Owner","aws_profile_name"]}',
            "prefix(\"aws_region\", 'bar') AND \"aws_region\" = 'us-east-1'",
            {"aws_region": "VARCHAR"},
        ),
        (
            '{"filters":[{"expression_class":"BOUND_OPERATOR","type":"COMPARE_IN","alias":"","query_location":81,"return_type":{"id":"BOOLEAN","type_info":null},"children":[{"expression_class":"BOUND_COLUMN_REF","type":"BOUND_COLUMN_REF","alias":"aws_region","query_location":70,"return_type":{"id":"VARCHAR","type_info":null},"binding":{"table_index":0,"column_index":0},"depth":0},{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"us-east-1"}},{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"us-east-2"}}]},{"expression_class":"BOUND_COMPARISON","type":"COMPARE_EQUAL","alias":"","query_location":18446744073709551615,"left":{"expression_class":"BOUND_COLUMN_REF","type":"BOUND_COLUMN_REF","alias":"aws_region","query_location":18446744073709551615,"return_type":{"id":"VARCHAR","type_info":null},"binding":{"table_index":0,"column_index":0},"depth":0},"right":{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"us-east-1"}}}],"column_binding_names_by_index":["aws_region","Buckets","Owner","aws_profile_name"]}',
            "\"aws_region\" IN ('us-east-1', 'us-east-2') AND \"aws_region\" = 'us-east-1'",
            {"aws_region": "VARCHAR"},
        ),
        (
            '{"filters":[{"expression_class":"BOUND_COMPARISON","type":"COMPARE_EQUAL","alias":"","query_location":18446744073709551615,"left":{"expression_class":"BOUND_FUNCTION","type":"BOUND_FUNCTION","alias":"","query_location":18446744073709551615,"return_type":{"id":"VARCHAR","type_info":null},"children":[{"expression_class":"BOUND_FUNCTION","type":"BOUND_FUNCTION","alias":"","query_location":18446744073709551615,"return_type":{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"Name","second":{"id":"VARCHAR","type_info":null}},{"first":"CreationDate","second":{"id":"TIMESTAMP_S","type_info":null}}]}},"children":[{"expression_class":"BOUND_COLUMN_REF","type":"BOUND_COLUMN_REF","alias":"Buckets","query_location":18446744073709551615,"return_type":{"id":"LIST","type_info":{"type":"LIST_TYPE_INFO","alias":"","modifiers":[],"child_type":{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"Name","second":{"id":"VARCHAR","type_info":null}},{"first":"CreationDate","second":{"id":"TIMESTAMP_S","type_info":null}}]}}}},"binding":{"table_index":0,"column_index":1},"depth":0},{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"BIGINT","type_info":null},"is_null":false,"value":1}}],"name":"array_extract","arguments":[{"id":"LIST","type_info":{"type":"LIST_TYPE_INFO","alias":"","modifiers":[],"child_type":{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"Name","second":{"id":"VARCHAR","type_info":null}},{"first":"CreationDate","second":{"id":"TIMESTAMP_S","type_info":null}}]}}}},{"id":"BIGINT","type_info":null}],"original_arguments":[],"has_serialize":false,"is_operator":false},{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"Name"}}],"name":"struct_extract","arguments":[{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"Name","second":{"id":"VARCHAR","type_info":null}},{"first":"CreationDate","second":{"id":"TIMESTAMP_S","type_info":null}}]}},{"id":"VARCHAR","type_info":null}],"original_arguments":[],"has_serialize":false,"is_operator":false},"right":{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"foobar"}}},{"expression_class":"BOUND_COMPARISON","type":"COMPARE_EQUAL","alias":"","query_location":18446744073709551615,"left":{"expression_class":"BOUND_COLUMN_REF","type":"BOUND_COLUMN_REF","alias":"aws_region","query_location":18446744073709551615,"return_type":{"id":"VARCHAR","type_info":null},"binding":{"table_index":0,"column_index":0},"depth":0},"right":{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"us-east-1"}}}],"column_binding_names_by_index":["aws_region","Buckets","Owner","aws_profile_name"]}',
            "struct_extract(array_extract(\"Buckets\", 1), 'Name') = 'foobar' AND \"aws_region\" = 'us-east-1'",
            bucket_type_info,
        ),
        (
            '{"filters":[{"expression_class":"BOUND_COMPARISON","type":"COMPARE_EQUAL","alias":"","query_location":18446744073709551615,"left":{"expression_class":"BOUND_FUNCTION","type":"BOUND_FUNCTION","alias":"","query_location":71,"return_type":{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"hello","second":{"id":"VARCHAR","type_info":null}}]}},"children":[{"expression_class":"BOUND_FUNCTION","type":"BOUND_FUNCTION","alias":"hello","query_location":18446744073709551615,"return_type":{"id":"VARCHAR","type_info":null},"children":[{"expression_class":"BOUND_FUNCTION","type":"BOUND_FUNCTION","alias":"","query_location":18446744073709551615,"return_type":{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"Name","second":{"id":"VARCHAR","type_info":null}},{"first":"CreationDate","second":{"id":"TIMESTAMP_S","type_info":null}}]}},"children":[{"expression_class":"BOUND_COLUMN_REF","type":"BOUND_COLUMN_REF","alias":"Buckets","query_location":18446744073709551615,"return_type":{"id":"LIST","type_info":{"type":"LIST_TYPE_INFO","alias":"","modifiers":[],"child_type":{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"Name","second":{"id":"VARCHAR","type_info":null}},{"first":"CreationDate","second":{"id":"TIMESTAMP_S","type_info":null}}]}}}},"binding":{"table_index":0,"column_index":1},"depth":0},{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"BIGINT","type_info":null},"is_null":false,"value":1}}],"name":"array_extract","arguments":[{"id":"LIST","type_info":{"type":"LIST_TYPE_INFO","alias":"","modifiers":[],"child_type":{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"Name","second":{"id":"VARCHAR","type_info":null}},{"first":"CreationDate","second":{"id":"TIMESTAMP_S","type_info":null}}]}}}},{"id":"BIGINT","type_info":null}],"original_arguments":[],"has_serialize":false,"is_operator":false},{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"Name"}}],"name":"struct_extract","arguments":[{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"Name","second":{"id":"VARCHAR","type_info":null}},{"first":"CreationDate","second":{"id":"TIMESTAMP_S","type_info":null}}]}},{"id":"VARCHAR","type_info":null}],"original_arguments":[],"has_serialize":false,"is_operator":false}],"name":"struct_pack","arguments":[],"original_arguments":[],"has_serialize":true,"function_data":{"variable_return_type":{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"hello","second":{"id":"VARCHAR","type_info":null}}]}}},"is_operator":false},"right":{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"hello","second":{"id":"VARCHAR","type_info":null}}]}},"is_null":false,"value":{"children":[{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"foo"}]}}}},{"expression_class":"BOUND_COMPARISON","type":"COMPARE_EQUAL","alias":"","query_location":18446744073709551615,"left":{"expression_class":"BOUND_COLUMN_REF","type":"BOUND_COLUMN_REF","alias":"aws_region","query_location":18446744073709551615,"return_type":{"id":"VARCHAR","type_info":null},"binding":{"table_index":0,"column_index":0},"depth":0},"right":{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"us-east-1"}}}],"column_binding_names_by_index":["aws_region","Buckets","Owner","aws_profile_name"]}',
            "struct_pack(hello := struct_extract(array_extract(\"Buckets\", 1), 'Name')) = {'hello':'foo'} AND \"aws_region\" = 'us-east-1'",
            bucket_type_info,
        ),
        (
            '{"filters":[{"expression_class":"BOUND_COMPARISON","type":"COMPARE_EQUAL","alias":"","query_location":18446744073709551615,"left":{"expression_class":"BOUND_FUNCTION","type":"BOUND_FUNCTION","alias":"","query_location":71,"return_type":{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"hello","second":{"id":"VARCHAR","type_info":null}},{"first":"d","second":{"id":"DATE","type_info":null}},{"first":"z","second":{"id":"LIST","type_info":{"type":"LIST_TYPE_INFO","alias":"","modifiers":[],"child_type":{"id":"INTEGER","type_info":null}}}},{"first":"c","second":{"id":"LIST","type_info":{"type":"LIST_TYPE_INFO","alias":"","modifiers":[],"child_type":{"id":"VARCHAR","type_info":null}}}}]}},"children":[{"expression_class":"BOUND_FUNCTION","type":"BOUND_FUNCTION","alias":"hello","query_location":18446744073709551615,"return_type":{"id":"VARCHAR","type_info":null},"children":[{"expression_class":"BOUND_FUNCTION","type":"BOUND_FUNCTION","alias":"","query_location":18446744073709551615,"return_type":{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"Name","second":{"id":"VARCHAR","type_info":null}},{"first":"CreationDate","second":{"id":"TIMESTAMP_S","type_info":null}}]}},"children":[{"expression_class":"BOUND_COLUMN_REF","type":"BOUND_COLUMN_REF","alias":"Buckets","query_location":18446744073709551615,"return_type":{"id":"LIST","type_info":{"type":"LIST_TYPE_INFO","alias":"","modifiers":[],"child_type":{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"Name","second":{"id":"VARCHAR","type_info":null}},{"first":"CreationDate","second":{"id":"TIMESTAMP_S","type_info":null}}]}}}},"binding":{"table_index":0,"column_index":1},"depth":0},{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"BIGINT","type_info":null},"is_null":false,"value":1}}],"name":"array_extract","arguments":[{"id":"LIST","type_info":{"type":"LIST_TYPE_INFO","alias":"","modifiers":[],"child_type":{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"Name","second":{"id":"VARCHAR","type_info":null}},{"first":"CreationDate","second":{"id":"TIMESTAMP_S","type_info":null}}]}}}},{"id":"BIGINT","type_info":null}],"original_arguments":[],"has_serialize":false,"is_operator":false},{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"Name"}}],"name":"struct_extract","arguments":[{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"Name","second":{"id":"VARCHAR","type_info":null}},{"first":"CreationDate","second":{"id":"TIMESTAMP_S","type_info":null}}]}},{"id":"VARCHAR","type_info":null}],"original_arguments":[],"has_serialize":false,"is_operator":false},{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"DATE","type_info":null},"is_null":false,"value":19754}},{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"LIST","type_info":{"type":"LIST_TYPE_INFO","alias":"","modifiers":[],"child_type":{"id":"INTEGER","type_info":null}}},"is_null":false,"value":{"children":[{"type":{"id":"INTEGER","type_info":null},"is_null":false,"value":1},{"type":{"id":"INTEGER","type_info":null},"is_null":false,"value":2},{"type":{"id":"INTEGER","type_info":null},"is_null":false,"value":3}]}}},{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"LIST","type_info":{"type":"LIST_TYPE_INFO","alias":"","modifiers":[],"child_type":{"id":"VARCHAR","type_info":null}}},"is_null":false,"value":{"children":[{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"f"},{"type":{"id":"VARCHAR","type_info":null},"is_null":true},{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"b"}]}}}],"name":"struct_pack","arguments":[],"original_arguments":[],"has_serialize":true,"function_data":{"variable_return_type":{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"hello","second":{"id":"VARCHAR","type_info":null}},{"first":"d","second":{"id":"DATE","type_info":null}},{"first":"z","second":{"id":"LIST","type_info":{"type":"LIST_TYPE_INFO","alias":"","modifiers":[],"child_type":{"id":"INTEGER","type_info":null}}}},{"first":"c","second":{"id":"LIST","type_info":{"type":"LIST_TYPE_INFO","alias":"","modifiers":[],"child_type":{"id":"VARCHAR","type_info":null}}}}]}}},"is_operator":false},"right":{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"hello","second":{"id":"VARCHAR","type_info":null}},{"first":"d","second":{"id":"DATE","type_info":null}},{"first":"z","second":{"id":"LIST","type_info":{"type":"LIST_TYPE_INFO","alias":"","modifiers":[],"child_type":{"id":"INTEGER","type_info":null}}}},{"first":"c","second":{"id":"LIST","type_info":{"type":"LIST_TYPE_INFO","alias":"","modifiers":[],"child_type":{"id":"VARCHAR","type_info":null}}}}]}},"is_null":false,"value":{"children":[{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"foo"},{"type":{"id":"DATE","type_info":null},"is_null":false,"value":3652},{"type":{"id":"LIST","type_info":{"type":"LIST_TYPE_INFO","alias":"","modifiers":[],"child_type":{"id":"INTEGER","type_info":null}}},"is_null":false,"value":{"children":[{"type":{"id":"INTEGER","type_info":null},"is_null":false,"value":4},{"type":{"id":"INTEGER","type_info":null},"is_null":false,"value":5},{"type":{"id":"INTEGER","type_info":null},"is_null":false,"value":6}]}},{"type":{"id":"LIST","type_info":{"type":"LIST_TYPE_INFO","alias":"","modifiers":[],"child_type":{"id":"VARCHAR","type_info":null}}},"is_null":false,"value":{"children":[{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"f"},{"type":{"id":"VARCHAR","type_info":null},"is_null":true},{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"b"}]}}]}}}},{"expression_class":"BOUND_COMPARISON","type":"COMPARE_EQUAL","alias":"","query_location":18446744073709551615,"left":{"expression_class":"BOUND_COLUMN_REF","type":"BOUND_COLUMN_REF","alias":"aws_region","query_location":18446744073709551615,"return_type":{"id":"VARCHAR","type_info":null},"binding":{"table_index":0,"column_index":0},"depth":0},"right":{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"us-east-1"}}}],"column_binding_names_by_index":["aws_region","Buckets","Owner","aws_profile_name"]}',
            "struct_pack(hello := struct_extract(array_extract(\"Buckets\", 1), 'Name'), d := '2024-02-01', z := [1, 2, 3], c := ['f', null, 'b']) = {'hello':'foo','d':'1980-01-01','z':[4, 5, 6],'c':['f', null, 'b']} AND \"aws_region\" = 'us-east-1'",
            bucket_type_info,
        ),
        (
            '{"filters":[{"expression_class":"BOUND_COMPARISON","type":"COMPARE_EQUAL","alias":"","query_location":78,"left":{"expression_class":"BOUND_FUNCTION","type":"BOUND_FUNCTION","alias":"","query_location":71,"return_type":{"id":"LIST","type_info":{"type":"LIST_TYPE_INFO","alias":"","modifiers":[],"child_type":{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"DisplayName","second":{"id":"VARCHAR","type_info":null}},{"first":"ID","second":{"id":"VARCHAR","type_info":null}}]}}}},"children":[{"expression_class":"BOUND_COLUMN_REF","type":"BOUND_COLUMN_REF","alias":"Owner","query_location":71,"return_type":{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"DisplayName","second":{"id":"VARCHAR","type_info":null}},{"first":"ID","second":{"id":"VARCHAR","type_info":null}}]}},"binding":{"table_index":0,"column_index":1},"depth":0}],"name":"list_value","arguments":[],"original_arguments":[],"has_serialize":false,"is_operator":false},"right":{"expression_class":"BOUND_FUNCTION","type":"BOUND_FUNCTION","alias":"","query_location":81,"return_type":{"id":"LIST","type_info":{"type":"LIST_TYPE_INFO","alias":"","modifiers":[],"child_type":{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"DisplayName","second":{"id":"VARCHAR","type_info":null}},{"first":"ID","second":{"id":"VARCHAR","type_info":null}}]}}}},"children":[{"expression_class":"BOUND_COLUMN_REF","type":"BOUND_COLUMN_REF","alias":"Owner","query_location":81,"return_type":{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"DisplayName","second":{"id":"VARCHAR","type_info":null}},{"first":"ID","second":{"id":"VARCHAR","type_info":null}}]}},"binding":{"table_index":0,"column_index":1},"depth":0}],"name":"list_value","arguments":[],"original_arguments":[],"has_serialize":false,"is_operator":false}},{"expression_class":"BOUND_COMPARISON","type":"COMPARE_EQUAL","alias":"","query_location":18446744073709551615,"left":{"expression_class":"BOUND_COLUMN_REF","type":"BOUND_COLUMN_REF","alias":"aws_region","query_location":18446744073709551615,"return_type":{"id":"VARCHAR","type_info":null},"binding":{"table_index":0,"column_index":0},"depth":0},"right":{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"us-east-1"}}}],"column_binding_names_by_index":["aws_region","Owner","Buckets","aws_profile_name"]}',
            'list_value("Owner") = list_value("Owner") AND "aws_region" = \'us-east-1\'',
            owner_type_info,
        ),
        (
            '{"filters":[{"expression_class":"BOUND_COMPARISON","type":"COMPARE_EQUAL","alias":"","query_location":18446744073709551615,"left":{"expression_class":"BOUND_COLUMN_REF","type":"BOUND_COLUMN_REF","alias":"Owner","query_location":18446744073709551615,"return_type":{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"DisplayName","second":{"id":"VARCHAR","type_info":null}},{"first":"ID","second":{"id":"VARCHAR","type_info":null}}]}},"binding":{"table_index":0,"column_index":1},"depth":0},"right":{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"DisplayName","second":{"id":"VARCHAR","type_info":null}},{"first":"ID","second":{"id":"VARCHAR","type_info":null}}]}},"is_null":false,"value":{"children":[{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"test"},{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"v"}]}}}},{"expression_class":"BOUND_COMPARISON","type":"COMPARE_EQUAL","alias":"","query_location":18446744073709551615,"left":{"expression_class":"BOUND_COLUMN_REF","type":"BOUND_COLUMN_REF","alias":"aws_region","query_location":18446744073709551615,"return_type":{"id":"VARCHAR","type_info":null},"binding":{"table_index":0,"column_index":0},"depth":0},"right":{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"us-east-1"}}}],"column_binding_names_by_index":["aws_region","Owner","Buckets","aws_profile_name"]}',
            "\"Owner\" = {'DisplayName':'test','ID':'v'} AND \"aws_region\" = 'us-east-1'",
            owner_type_info,
        ),
        (
            '{"filters":[{"expression_class":"BOUND_OPERATOR","type":"COMPARE_NOT_IN","alias":"","query_location":76,"return_type":{"id":"BOOLEAN","type_info":null},"children":[{"expression_class":"BOUND_COLUMN_REF","type":"BOUND_COLUMN_REF","alias":"Owner","query_location":70,"return_type":{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"DisplayName","second":{"id":"VARCHAR","type_info":null}},{"first":"ID","second":{"id":"VARCHAR","type_info":null}}]}},"binding":{"table_index":0,"column_index":1},"depth":0},{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"DisplayName","second":{"id":"VARCHAR","type_info":null}},{"first":"ID","second":{"id":"VARCHAR","type_info":null}}]}},"is_null":false,"value":{"children":[{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"test"},{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"v"}]}}},{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"DisplayName","second":{"id":"VARCHAR","type_info":null}},{"first":"ID","second":{"id":"VARCHAR","type_info":null}}]}},"is_null":false,"value":{"children":[{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"test2"},{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"v2"}]}}}]},{"expression_class":"BOUND_COMPARISON","type":"COMPARE_EQUAL","alias":"","query_location":18446744073709551615,"left":{"expression_class":"BOUND_COLUMN_REF","type":"BOUND_COLUMN_REF","alias":"aws_region","query_location":18446744073709551615,"return_type":{"id":"VARCHAR","type_info":null},"binding":{"table_index":0,"column_index":0},"depth":0},"right":{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"us-east-1"}}}],"column_binding_names_by_index":["aws_region","Owner","Buckets","aws_profile_name"]}',
            "\"Owner\" NOT IN ({'DisplayName':'test','ID':'v'}, {'DisplayName':'test2','ID':'v2'}) AND \"aws_region\" = 'us-east-1'",
            owner_type_info,
        ),
        (
            '{"filters":[{"expression_class":"BOUND_BETWEEN","type":"COMPARE_BETWEEN","alias":"","query_location":18446744073709551615,"input":{"expression_class":"BOUND_FUNCTION","type":"BOUND_FUNCTION","alias":"","query_location":70,"return_type":{"id":"VARCHAR","type_info":null},"children":[{"expression_class":"BOUND_COLUMN_REF","type":"BOUND_COLUMN_REF","alias":"Owner","query_location":18446744073709551615,"return_type":{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"DisplayName","second":{"id":"VARCHAR","type_info":null}},{"first":"ID","second":{"id":"VARCHAR","type_info":null}}]}},"binding":{"table_index":0,"column_index":1},"depth":0},{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"DisplayName"}}],"name":"struct_extract","arguments":[{"id":"STRUCT","type_info":{"type":"STRUCT_TYPE_INFO","alias":"","modifiers":[],"child_types":[{"first":"DisplayName","second":{"id":"VARCHAR","type_info":null}},{"first":"ID","second":{"id":"VARCHAR","type_info":null}}]}},{"id":"VARCHAR","type_info":null}],"original_arguments":[],"has_serialize":false,"is_operator":false},"lower":{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"a"}},"upper":{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"z"}},"lower_inclusive":true,"upper_inclusive":true},{"expression_class":"BOUND_COMPARISON","type":"COMPARE_EQUAL","alias":"","query_location":18446744073709551615,"left":{"expression_class":"BOUND_COLUMN_REF","type":"BOUND_COLUMN_REF","alias":"aws_region","query_location":18446744073709551615,"return_type":{"id":"VARCHAR","type_info":null},"binding":{"table_index":0,"column_index":0},"depth":0},"right":{"expression_class":"BOUND_CONSTANT","type":"VALUE_CONSTANT","alias":"","query_location":18446744073709551615,"value":{"type":{"id":"VARCHAR","type_info":null},"is_null":false,"value":"us-east-1"}}}],"column_binding_names_by_index":["aws_region","Owner","Buckets","aws_profile_name"]}',
            "struct_extract(\"Owner\", 'DisplayName') BETWEEN 'a' AND 'z' AND \"aws_region\" = 'us-east-1'",
            owner_type_info,
        ),
    ],
)
def test_duckdb_json_to_sql(
    json_source: str, expected_sql: str, expected_bound_types: dict[str, Any]
) -> None:
    data = json.loads(json_source)
    result = expression.convert_to_sql(
        source=data["filters"], bound_column_names=data["column_binding_names_by_index"]
    )
    assert result[0] == expected_sql
    assert result[1] == expected_bound_types



import base64
import math
import uuid
from datetime import UTC, date, datetime, time, timedelta
from decimal import Decimal
from typing import Annotated, Any, Literal

from pydantic import BaseModel, Discriminator, Tag, field_validator


class ValueBase(BaseModel):
    is_null: bool


class ValueValue_128bits(BaseModel):
    """
    Represent a 128-bit value with an upper and lower half.
    """

    lower: int
    upper: int


def decode_base64(cls: Any, value: Any) -> bytes:
    """
    Decode a Base64 encoded value.
    """
    try:
        return base64.b64decode(value)
    except Exception as e:
        raise ValueError(f"Invalid Base64 encoded value: {e}") from e


class ValueValue_base64(BaseModel):
    """
    Store and decode a base64 value
    """

    base64: bytes
    # Do the decoding of the value here for the user.
    _validate_base64 = field_validator("base64", mode="before")(decode_base64)


class ValueTypeInfo_list(BaseModel):
    alias: str
    child_type: "AnyValueType"
    type: Literal["LIST_TYPE_INFO"] = "LIST_TYPE_INFO"


class ValueType_list(BaseModel):
    id: Literal["LIST"] = "LIST"
    type_info: ValueTypeInfo_list

    def sql(self) -> str:
        return f"{self.type_info.child_type.sql()}[]"


class ValueValue_list(BaseModel):
    children: list["Value"]


class Value_list(ValueBase):
    type: ValueType_list
    value: ValueValue_list

    def sql(self) -> str:
        return "[" + ", ".join([child.sql() for child in self.value.children]) + "]"


class ValueType_bigint(BaseModel):
    id: Literal["BIGINT"] = "BIGINT"
    type_info: None

    def sql(self) -> str:
        return self.id


class Value_bigint(ValueBase):
    type: ValueType_bigint
    value: int | None = None

    def sql(self) -> str:
        if self.is_null:
            return "null"
        return str(self.value)


class ValueType_bit(BaseModel):
    id: Literal["BIT"] = "BIT"
    type_info: None

    def sql(self) -> str:
        return self.id


class Value_bit(ValueBase):
    type: ValueType_bit
    value: bytes | ValueValue_base64 | None = None

    def sql(self) -> str:
        if self.is_null:
            return "null"
        assert self.value

        data = (
            self.value.base64
            if isinstance(self.value, ValueValue_base64)
            else self.value
        )

        if not data or len(data) < 2:
            return ""

        padding_bits = data[0]
        bit_data = data[1:]

        # Convert all bytes to bits
        bits = "".join(f"{byte:08b}" for byte in bit_data)

        # Remove the padding bits from the end
        if padding_bits:
            bits = bits[padding_bits:]

        return f"'{bits}'"


class ValueType_blob(BaseModel):
    id: Literal["BLOB"] = "BLOB"
    type_info: None

    def sql(self) -> str:
        return self.id


class Value_blob(ValueBase):
    type: ValueType_blob
    value: str | None = None

    def sql(self) -> str:
        if self.is_null:
            return "null"
        assert self.value

        return "'" + self.value.replace("'", "''") + "'"


class ValueType_boolean(BaseModel):
    id: Literal["BOOLEAN"] = "BOOLEAN"
    type_info: None

    def sql(self) -> str:
        return self.id


class ValueType_any(BaseModel):
    id: Literal["ANY"] = "ANY"
    type_info: None

    def sql(self) -> str:
        raise NotImplementedError("ValueType_any.sql() is not implemented.")


class Value_boolean(ValueBase):
    type: ValueType_boolean
    value: bool | None = None

    def sql(self) -> str:
        if self.is_null:
            return "null"

        return str("true" if self.value else "false")


class ValueType_date(BaseModel):
    id: Literal["DATE"] = "DATE"
    type_info: None

    def sql(self) -> str:
        return self.id


class Value_date(ValueBase):
    type: ValueType_date
    value: int | None = None

    def sql(self) -> str:
        if self.is_null:
            return "null"
        assert self.value

        if self.value == -2147483647:
            return "'-infinity'"
        elif self.value == 2147483647:
            return "'infinity'"
        formatted_date = (date(1970, 1, 1) + timedelta(days=self.value)).isoformat()
        return f"'{formatted_date}'"


class ValueTypeInfo_decimal(BaseModel):
    width: int
    scale: int


class ValueType_decimal(BaseModel):
    id: Literal["DECIMAL"] = "DECIMAL"
    type_info: ValueTypeInfo_decimal

    def sql(self) -> str:
        return f"DECIMAL({self.type_info.width}, {self.type_info.scale})"


class Value_decimal(ValueBase):
    type: ValueType_decimal
    value: int | ValueValue_128bits | None = None

    def sql(self) -> str:
        if self.is_null:
            return "null"
        assert self.value

        scale = self.type.type_info.scale

        if isinstance(self.value, ValueValue_128bits):
            # Reconstruct full integer (assuming 64-bit halves)
            combined = (self.value.upper << 64) | self.value.lower

            # Convert from unsigned to signed (two's complement if necessary)
            if self.value.upper & (1 << 63):
                combined -= 1 << 128

            decimal_value = Decimal(combined)
        elif isinstance(self.value, int):
            # Assume it's a simple int (64-bit)
            decimal_value = Decimal(self.value)
        else:
            raise NotImplementedError("Invalid Decimal value storage")

        return str(decimal_value / Decimal(10) ** scale)


class ValueType_double(BaseModel):
    id: Literal["DOUBLE"] = "DOUBLE"
    type_info: None

    def sql(self) -> str:
        return self.id


class Value_double(ValueBase):
    type: ValueType_double
    value: float | None = None

    def sql(self) -> str:
        if self.is_null:
            return "null"
        assert self.value

        if math.isinf(self.value):
            if self.value > 0:
                return "'infinity'"
            return "'-infinity'"
        elif math.isnan(self.value):
            return "'nan'"
        return str(self.value)


class ValueType_float(BaseModel):
    id: Literal["FLOAT"] = "FLOAT"
    type_info: None

    def sql(self) -> str:
        return self.id


class Value_float(ValueBase):
    type: ValueType_float
    value: float | None = None

    def sql(self) -> str:
        if self.is_null:
            return "null"
        assert self.value

        if math.isinf(self.value):
            if self.value > 0:
                return "'infinity'"
            return "'-infinity'"
        elif math.isnan(self.value):
            return "'nan'"
        return str(self.value)


class ValueType_hugeint(BaseModel):
    id: Literal["HUGEINT"] = "HUGEINT"
    type_info: None

    def sql(self) -> str:
        return self.id


class Value_hugeint(ValueBase):
    type: ValueType_hugeint
    value: ValueValue_128bits | None = None

    def sql(self) -> str:
        if self.is_null:
            return "null"
        assert self.value

        upper = self.value.upper
        lower = self.value.lower
        result = (upper << 64) | lower

        # If the highest bit (bit 127) is set, interpret as negative
        if upper & (1 << 63):
            result -= 1 << 128

        return str(result)


class ValueType_integer(BaseModel):
    id: Literal["INTEGER"] = "INTEGER"
    type_info: None

    def sql(self) -> str:
        return self.id


class Value_integer(ValueBase):
    type: ValueType_integer
    value: int | None = None

    def sql(self) -> str:
        if self.is_null:
            return "null"

        return str(self.value)


class ValueType_interval(BaseModel):
    id: Literal["INTERVAL"] = "INTERVAL"
    type_info: None

    def sql(self) -> str:
        return self.id


class ValueValue_interval(BaseModel):
    months: int
    days: int
    micros: int


class Value_interval(ValueBase):
    type: ValueType_interval
    value: ValueValue_interval | None = None

    def sql(self) -> str:
        if self.is_null:
            return "null"
        assert self.value

        return (
            "INTERVAL '"
            + f"{self.value.months} months {self.value.days} days {self.value.micros} us"
            + "'"
        )


class ValueType_null(BaseModel):
    id: Literal["NULL"] = "NULL"
    type_info: None

    def sql(self) -> str:
        return self.id


class Value_null(ValueBase):
    type: ValueType_null
    value: None

    def sql(self) -> str:
        return "null"


class ValueType_smallint(BaseModel):
    id: Literal["SMALLINT"] = "SMALLINT"
    type_info: None

    def sql(self) -> str:
        return self.id


class Value_smallint(ValueBase):
    type: ValueType_smallint
    value: int | None = None

    def sql(self) -> str:
        if self.is_null:
            return "null"

        return str(self.value)


class ValueType_time(BaseModel):
    id: Literal["TIME"] = "TIME"
    type_info: None

    def sql(self) -> str:
        return self.id


class Value_time(ValueBase):
    type: ValueType_time
    value: int | None = None

    def sql(self) -> str:
        if self.is_null:
            return "null"
        assert self.value

        t = timedelta(microseconds=self.value)
        hours, remainder = divmod(t.seconds, 3600)
        minutes, seconds = divmod(remainder, 60)

        return (
            "TIME '"
            + time(hours, minutes, seconds, microsecond=t.microseconds).strftime(
                "%H:%M:%S.%f"
            )
            + "'"
        )


class ValueType_time_with_time_zone(BaseModel):
    id: Literal["TIME WITH TIME ZONE"] = "TIME WITH TIME ZONE"
    type_info: None

    def sql(self) -> str:
        return self.id


class Value_time_with_time_zone(ValueBase):
    type: ValueType_time_with_time_zone
    value: int | None = None

    def sql(self) -> str:
        if self.is_null:
            return "null"
        assert self.value

        t = timedelta(microseconds=self.value)
        hours, remainder = divmod(t.seconds, 3600)
        minutes, seconds = divmod(remainder, 60)

        return (
            "TIMETZ '"
            + time(hours, minutes, seconds, microsecond=t.microseconds).strftime(
                "%H:%M:%S.%f"
            )
            + "'"
        )


class ValueType_timestamp_with_time_zone(BaseModel):
    id: Literal["TIMESTAMP WITH TIME ZONE"] = "TIMESTAMP WITH TIME ZONE"
    type_info: None

    def sql(self) -> str:
        return self.id


class Value_timestamp_with_time_zone(ValueBase):
    type: ValueType_timestamp_with_time_zone
    value: int | None = None

    def sql(self) -> str:
        if self.is_null:
            return "null"
        assert self.value

        return (
            "TIMESTAMPTZ '"
            + datetime.fromtimestamp(int(self.value) / 1_000_000, tz=UTC).strftime(
                "%Y-%m-%d %H:%M:%S.%f"
            )
            + "'"
        )


class ValueType_timestamp(BaseModel):
    id: Literal["TIMESTAMP"] = "TIMESTAMP"
    type_info: None

    def sql(self) -> str:
        return self.id


class Value_timestamp(ValueBase):
    type: ValueType_timestamp
    value: int | None = None

    def sql(self) -> str:
        if self.is_null:
            return "null"
        assert self.value

        return (
            "TIMESTAMP '"
            + datetime.fromtimestamp(int(self.value) / 1_000_000, tz=UTC).strftime(
                "%Y-%m-%d %H:%M:%S.%f"
            )
            + "'"
        )


class ValueType_timestamp_ms(BaseModel):
    id: Literal["TIMESTAMP_MS"] = "TIMESTAMP_MS"
    type_info: None

    def sql(self) -> str:
        return self.id


class Value_timestamp_ms(ValueBase):
    type: ValueType_timestamp_ms
    value: int | None = None

    def sql(self) -> str:
        if self.is_null:
            return "null"
        assert self.value

        dt = datetime.fromtimestamp(self.value / 1000, tz=UTC)
        return "TIMESTAMP_MS '" + dt.strftime("%Y-%m-%d %H:%M:%S.%f")[:-3] + "'"


class ValueType_timestamp_ns(BaseModel):
    id: Literal["TIMESTAMP_NS"] = "TIMESTAMP_NS"
    type_info: None

    def sql(self) -> str:
        return self.id


class Value_timestamp_ns(ValueBase):
    type: ValueType_timestamp_ns
    value: int | None = None

    def sql(self) -> str:
        if self.is_null:
            return "null"
        assert self.value
        ns_since_epoch = self.value
        seconds, nanoseconds = divmod(ns_since_epoch, 10**9)

        dt = datetime.fromtimestamp(seconds, tz=UTC)
        formatted = dt.strftime("%Y-%m-%dT%H:%M:%S") + f".{nanoseconds:09d}"
        return f"TIMESTAMP_NS '{formatted}'"


class ValueType_timestamp_s(BaseModel):
    id: Literal["TIMESTAMP_S"] = "TIMESTAMP_S"
    type_info: None

    def sql(self) -> str:
        return self.id


class Value_timestamp_s(ValueBase):
    type: ValueType_timestamp_s
    value: int | None = None

    def sql(self) -> str:
        if self.is_null:
            return "null"
        assert self.value

        dt = datetime.fromtimestamp(self.value, tz=UTC)
        return "TIMESTAMP_S '" + dt.strftime("%Y-%m-%d %H:%M:%S") + "'"


class ValueType_tinyint(BaseModel):
    id: Literal["TINYINT"] = "TINYINT"
    type_info: None

    def sql(self) -> str:
        return self.id


class Value_tinyint(ValueBase):
    type: ValueType_tinyint
    value: int | None = None

    def sql(self) -> str:
        if self.is_null:
            return "null"

        return str(self.value)


class ValueType_ubigint(BaseModel):
    id: Literal["UBIGINT"] = "UBIGINT"
    type_info: None

    def sql(self) -> str:
        return self.id


class Value_ubigint(ValueBase):
    type: ValueType_ubigint
    value: int | None = None

    def sql(self) -> str:
        if self.is_null:
            return "null"

        return str(self.value)


class ValueType_uhugeint(BaseModel):
    id: Literal["UHUGEINT"] = "UHUGEINT"
    type_info: None

    def sql(self) -> str:
        return self.id


class Value_uhugeint(ValueBase):
    type: ValueType_uhugeint
    value: ValueValue_128bits | None = None

    def sql(self) -> str:
        if self.is_null:
            return "null"
        assert self.value

        upper = self.value.upper
        lower = self.value.lower
        return str((upper << 64) | lower)


class ValueType_uinteger(BaseModel):
    id: Literal["UINTEGER"] = "UINTEGER"
    type_info: None

    def sql(self) -> str:
        return self.id


class Value_uinteger(ValueBase):
    type: ValueType_uinteger
    value: int | None = None

    def sql(self) -> str:
        if self.is_null:
            return "null"
        assert self.value
        return str(self.value)


class ValueType_usmallint(BaseModel):
    id: Literal["USMALLINT"] = "USMALLINT"
    type_info: None

    def sql(self) -> str:
        return self.id


class Value_usmallint(ValueBase):
    type: ValueType_usmallint
    value: int | None = None

    def sql(self) -> str:
        if self.is_null:
            return "null"
        assert self.value
        return str(self.value)


class ValueType_utinyint(BaseModel):
    id: Literal["UTINYINT"] = "UTINYINT"
    type_info: None

    def sql(self) -> str:
        return self.id


class Value_utinyint(ValueBase):
    type: ValueType_utinyint
    value: int | None = None

    def sql(self) -> str:
        if self.is_null:
            return "null"
        assert self.value
        return str(self.value)


class ValueType_uuid(BaseModel):
    id: Literal["UUID"] = "UUID"
    type_info: None

    def sql(self) -> str:
        return self.id


class Value_uuid(ValueBase):
    type: ValueType_uuid
    value: ValueValue_128bits | None = None

    def sql(self) -> str:
        if self.is_null:
            return "null"
        assert self.value
        upper = self.value.upper & ((1 << 64) - 1)  # Convert to unsigned if needed
        lower = self.value.lower

        # Combine into 128-bit integer
        combined = (upper << 64) | lower

        # Convert to 16 bytes (big-endian)
        bytes_ = combined.to_bytes(16, byteorder="big")

        # Create UUID from bytes
        u = uuid.UUID(bytes=bytes_)

        return str(u)


class ValueType_varchar(BaseModel):
    id: Literal["VARCHAR"] = "VARCHAR"
    type_info: None

    def sql(self) -> str:
        return self.id


class Value_varchar(ValueBase):
    type: ValueType_varchar
    value: str | None = None

    def sql(self) -> str:
        if self.is_null:
            return "null"
        assert self.value

        return "'" + self.value.replace("'", "''") + "'"


class ValueType_bignum(BaseModel):
    id: Literal["BIGNUM"] = "BIGNUM"
    type_info: None

    def sql(self) -> str:
        return self.id


def _bignum_get_byte_array(blob: bytes) -> tuple[list[int], bool]:
    if len(blob) < 4:
        raise ValueError("Invalid blob size.")

    # Determine if the number is negative
    is_negative = (blob[0] & 0x80) == 0

    # Extract byte array starting from the 4th byte
    byte_array = [~b & 255 for b in blob[3:]] if is_negative else list(blob[3:])
    return byte_array, is_negative


class Value_bignum(ValueBase):
    type: ValueType_bignum
    value: bytes | ValueValue_base64 | None = None

    def sql(self) -> str:
        if self.is_null:
            return "null"
        assert self.value

        decimal_string = ""
        if isinstance(self.value, ValueValue_base64):
            byte_array, is_negative = _bignum_get_byte_array(self.value.base64)
        else:
            byte_array, is_negative = _bignum_get_byte_array(self.value)

        digits: list[int] = []

        # Constants matching your C++ code (update if needed)
        DIGIT_BYTES = 4  # Assuming 4 bytes per digit (like a uint32_t)
        DIGIT_BITS = 32
        DECIMAL_BASE = 1000000000  # Typically 10^9 for efficient base conversion
        DECIMAL_SHIFT = 9  # Number of decimal digits in DECIMAL_BASE

        # Pad the byte array so we can process in DIGIT_BYTES chunks without conditionals
        padding_size = (-len(byte_array)) & (DIGIT_BYTES - 1)
        byte_array = [0] * padding_size + byte_array

        for i in range(0, len(byte_array), DIGIT_BYTES):
            hi = 0
            for j in range(DIGIT_BYTES):
                hi |= byte_array[i + j] << (8 * (DIGIT_BYTES - j - 1))

            for j in range(len(digits)):
                tmp = (digits[j] << DIGIT_BITS) | hi
                hi = tmp // DECIMAL_BASE
                digits[j] = tmp - DECIMAL_BASE * hi

            while hi:
                digits.append(hi % DECIMAL_BASE)
                hi //= DECIMAL_BASE

        if not digits:
            digits.append(0)

        for i in range(len(digits) - 1):
            remain = digits[i]
            for _ in range(DECIMAL_SHIFT):
                decimal_string += str(remain % 10)
                remain //= 10

        remain = digits[-1]
        while remain != 0:
            decimal_string += str(remain % 10)
            remain //= 10

        if is_negative:
            decimal_string += "-"

        # Reverse the string to get the correct number
        decimal_string = decimal_string[::-1]
        return decimal_string if decimal_string else "0"


def get_discriminator_value(v: Any) -> str:
    return v.get("type").get("id")


class ValueTypeInfoChild_struct(BaseModel):
    first: str
    second: "AnyValueType"


class ValueTypeInfo_struct(BaseModel):
    child_types: list[ValueTypeInfoChild_struct]
    type: Literal["STRUCT_TYPE_INFO"] = "STRUCT_TYPE_INFO"
    alias: str | None = None


class ValueType_struct(BaseModel):
    id: Literal["STRUCT"] = "STRUCT"
    type_info: ValueTypeInfo_struct

    def sql(self) -> str:
        return (
            "STRUCT("
            + ",".join(
                [
                    f'"{child.first}" {child.second.sql()}'
                    for child in self.type_info.child_types
                ]
            )
            + ")"
        )


class ValueValue_struct(BaseModel):
    children: list["Value"]


class Value_struct(ValueBase):
    type: ValueType_struct
    value: ValueValue_struct

    def sql(self) -> str:
        names = [child.first for child in self.type.type_info.child_types]
        values = self.value.children
        return (
            "{"
            + ",".join(
                [
                    f"'{name}':" + value.sql()
                    for name, value in zip(names, values, strict=True)
                ]
            )
            + "}"
        )


class ValueTypeInfo_map(BaseModel):
    child_type: ValueType_struct
    type: Literal["LIST_TYPE_INFO"] = "LIST_TYPE_INFO"
    alias: str | None = None


class ValueType_map(BaseModel):
    id: Literal["MAP"] = "MAP"
    type_info: ValueTypeInfo_map

    def sql(self) -> str:
        return (
            "MAP("
            + ",".join(
                [
                    f"{child.second.sql()}"
                    for child in self.type_info.child_type.type_info.child_types
                ]
            )
            + ")"
        )


class ValueValue_map(BaseModel):
    children: list["Value"]


class Value_map(ValueBase):
    type: ValueType_map
    value: ValueValue_map

    def sql(self) -> str:
        pairs: list[str] = []
        for child in self.value.children:
            assert isinstance(child.value, ValueValue_struct)
            k, v = child.value.children
            pairs.append(f"{k.sql()}:{v.sql()}")

        return "MAP {" + ",".join(pairs) + "}"


AnyValueType = (
    ValueType_any
    | ValueType_boolean
    | ValueType_bigint
    | ValueType_bit
    | ValueType_blob
    | ValueType_date
    | ValueType_decimal
    | ValueType_double
    | ValueType_float
    | ValueType_hugeint
    | ValueType_integer
    | ValueType_interval
    | ValueType_list
    | ValueType_map
    | ValueType_null
    | ValueType_smallint
    | ValueType_struct
    | ValueType_time
    | ValueType_time_with_time_zone
    | ValueType_timestamp
    | ValueType_timestamp_with_time_zone
    | ValueType_timestamp_ms
    | ValueType_timestamp_ns
    | ValueType_timestamp_s
    | ValueType_tinyint
    | ValueType_ubigint
    | ValueType_uhugeint
    | ValueType_uinteger
    | ValueType_usmallint
    | ValueType_utinyint
    | ValueType_uuid
    | ValueType_varchar
    | ValueType_bignum
)


Value = Annotated[
    Annotated[Value_bigint, Tag("BIGINT")]
    | Annotated[Value_bit, Tag("BIT")]
    | Annotated[Value_blob, Tag("BLOB")]
    | Annotated[Value_boolean, Tag("BOOLEAN")]
    | Annotated[Value_date, Tag("DATE")]
    | Annotated[Value_decimal, Tag("DECIMAL")]
    | Annotated[Value_double, Tag("DOUBLE")]
    | Annotated[Value_float, Tag("FLOAT")]
    | Annotated[Value_hugeint, Tag("HUGEINT")]
    | Annotated[Value_integer, Tag("INTEGER")]
    | Annotated[Value_interval, Tag("INTERVAL")]
    | Annotated[Value_list, Tag("LIST")]
    | Annotated[Value_map, Tag("MAP")]
    | Annotated[Value_null, Tag("NULL")]
    | Annotated[Value_smallint, Tag("SMALLINT")]
    | Annotated[Value_struct, Tag("STRUCT")]
    | Annotated[Value_time, Tag("TIME")]
    | Annotated[Value_time_with_time_zone, Tag("TIME WITH TIME ZONE")]
    | Annotated[Value_timestamp, Tag("TIMESTAMP")]
    | Annotated[Value_timestamp_with_time_zone, Tag("TIMESTAMP WITH TIME ZONE")]
    | Annotated[Value_timestamp_ms, Tag("TIMESTAMP_MS")]
    | Annotated[Value_timestamp_ns, Tag("TIMESTAMP_NS")]
    | Annotated[Value_timestamp_s, Tag("TIMESTAMP_S")]
    | Annotated[Value_tinyint, Tag("TINYINT")]
    | Annotated[Value_ubigint, Tag("UBIGINT")]
    | Annotated[Value_uhugeint, Tag("UHUGEINT")]
    | Annotated[Value_uinteger, Tag("UINTEGER")]
    | Annotated[Value_usmallint, Tag("USMALLINT")]
    | Annotated[Value_utinyint, Tag("UTINYINT")]
    | Annotated[Value_uuid, Tag("UUID")]
    | Annotated[Value_varchar, Tag("VARCHAR")]
    | Annotated[Value_bignum, Tag("BIGNUM")],
    Discriminator(get_discriminator_value),
]



from typing import Annotated, Literal, Union

from pydantic import BaseModel, Discriminator, Field, Tag

from .value import AnyValueType, Value, ValueType_struct


class Context(BaseModel):
    """
    Context for parsing expressions, containing information about bound columnns,
    and storing information about the types bound.
    """

    bound_column_names: list[str]
    bound_column_types: dict[str, str]


class BaseWithContext(BaseModel):
    """
    Base class for objects that can have a parse context injected.
    """

    parse_context_: Context | None = Field(default=None, exclude=True)


def inject_context(obj: BaseWithContext, context: Context) -> None:
    """
    Inject the context into the object and its nested BaseWithContext objects,
    this is needed for table and column references to be properly resolved.
    """
    if isinstance(obj, BaseWithContext):
        obj.parse_context_ = context
        for _field_name, field_value in obj.__dict__.items():
            if isinstance(field_value, BaseWithContext):
                inject_context(field_value, context)
            elif isinstance(field_value, list):
                for item in field_value:
                    inject_context(item, context)
            elif isinstance(field_value, dict):
                for item in field_value.values():
                    inject_context(item, context)


class Base(BaseWithContext):
    """
    Base class for all expressions
    """

    alias: str | None = None

    # The location of the expression in the query, used for error reporting.
    query_location: int | None = None


class BoundConstant(Base):
    """
    Represents a constant value in an expression.
    """

    expression_class: Literal["BOUND_CONSTANT"] = "BOUND_CONSTANT"
    type: Literal["VALUE_CONSTANT"] = "VALUE_CONSTANT"
    value: Value

    def sql(self) -> str:
        return self.value.sql()


class BoundColumnRefBinding(BaseModel):
    table_index: int
    column_index: int


class BoundColumnRef(Base, BaseWithContext):
    expression_class: Literal["BOUND_COLUMN_REF"] = "BOUND_COLUMN_REF"
    binding: BoundColumnRefBinding
    return_type: AnyValueType

    def sql(self) -> str:
        assert self.parse_context_
        column_name = self.parse_context_.bound_column_names[self.binding.column_index]
        self.parse_context_.bound_column_types[column_name] = self.return_type.sql()
        return f'"{column_name}"'


comparison_type_to_sql_operator: dict[str, str] = {
    "COMPARE_EQUAL": "=",
    "COMPARE_NOTEQUAL": "!=",
    "COMPARE_LESSTHAN": "<",
    "COMPARE_GREATERTHAN": ">",
    "COMPARE_LESSTHANOREQUALTO": "<=",
    "COMPARE_GREATERTHANOREQUALTO": ">=",
    "COMPARE_DISTINCT_FROM": "IS DISTINCT FROM",
    "COMPARE_NOT_DISTINCT_FROM": "IS NOT DISTINCT FROM",
}

Any = Union[
    BoundConstant,
    BoundColumnRef,
    "BoundComparison",
    "BoundCast",
    "BoundFunction",
    "BoundOperator",
    "BoundCase",
    "BoundBetween",
    "BoundConjunction",
]


class BoundConjunction(Base):
    """
    Represents a conjunction (AND/OR) of multiple expressions.
    """

    expression_class: Literal["BOUND_CONJUNCTION"] = "BOUND_CONJUNCTION"
    type: Literal["CONJUNCTION_AND", "CONJUNCTION_OR"]
    children: list[Any]

    def sql(self) -> str:
        operator = "AND" if self.type == "CONJUNCTION_AND" else "OR"
        return "(" + f" {operator} ".join([child.sql() for child in self.children]) + ")"


class BoundBetween(Base):
    """
    Represents a BETWEEN expression, which checks if a value is within a range.
    """

    expression_class: Literal["BOUND_BETWEEN"] = "BOUND_BETWEEN"
    input: Any
    lower: Any
    upper: Any

    def sql(self) -> str:
        return f"{self.input.sql()} BETWEEN {self.lower.sql()} AND {self.upper.sql()}"


class BoundCast(Base):
    """
    Represents a CAST expression, which converts a value from one type to another.
    """

    expression_class: Literal["BOUND_CAST"] = "BOUND_CAST"
    child: Any
    return_type: AnyValueType

    def sql(self) -> str:
        return f"CAST({self.child.sql()} AS {self.return_type.sql()})"


class BoundCaseCaseCheck(BaseModel):
    when_expr: Any
    then_expr: Any


class BoundCase(Base):
    """
    Represents a CASE expression, which evaluates a series of conditions and returns
    a corresponding value based on the first true condition.
    """

    expression_class: Literal["BOUND_CASE"] = "BOUND_CASE"
    case_checks: list[BoundCaseCaseCheck]
    else_expr: Any | None = None

    def sql(self) -> str:
        case_checks = [
            f"WHEN {case_check.when_expr.sql()} THEN {case_check.then_expr.sql()}"
            for case_check in self.case_checks
        ]
        if self.else_expr:
            case_checks.append(f"ELSE {self.else_expr.sql()}")

        return "CASE " + " ".join(case_checks) + " END"


class BoundOperator(Base):
    """
    Represents an operator expression, which can be a unary or binary operation,
    or a comparison operation.
    """

    expression_class: Literal["BOUND_OPERATOR"] = "BOUND_OPERATOR"
    type: Literal[
        "OPERATOR_IS_NULL",
        "OPERATOR_IS_NOT_NULL",
        "COMPARE_IN",
        "COMPARE_NOT_IN",
        "OPERATOR_NOT",
    ]
    children: list[Any]

    def sql(self) -> str:
        if self.type in ("OPERATOR_IS_NULL", "OPERATOR_IS_NOT_NULL"):
            operation = "IS NULL" if self.type == "OPERATOR_IS_NULL" else "IS NOT NULL"
            return self.children[0].sql() + " " + operation
        elif self.type in ("COMPARE_IN", "COMPARE_NOT_IN"):
            first, *rest = self.children
            operation = "IN" if self.type == "COMPARE_IN" else "NOT IN"
            return f"{first.sql()} {operation} ({', '.join([child.sql() for child in rest])})"
        elif self.type == "OPERATOR_NOT":
            assert len(self.children) == 1
            return f"NOT {self.children[0].sql()}"
        else:
            raise ValueError(f"Unsupported operator type: {self.type}")


class BoundFunctionFunctionData(BaseModel):
    variable_return_type: AnyValueType


class BoundFunction(Base):
    """
    Represents a function call in an expression, which can include user-defined functions,
    built-in functions, and operators.
    """

    expression_class: Literal["BOUND_FUNCTION"] = "BOUND_FUNCTION"
    type: Literal["BOUND_FUNCTION"] = "BOUND_FUNCTION"
    name: str
    return_type: AnyValueType
    children: list[Any]
    arguments: list[AnyValueType] | None = None
    original_arguments: list[AnyValueType] | None = None
    has_serialize: bool
    is_operator: bool
    function_data: BoundFunctionFunctionData | None = None

    def sql(self) -> str:
        if self.name == "struct_pack":
            assert self.function_data is not None
            result_struct = self.function_data.variable_return_type
            assert isinstance(result_struct, ValueType_struct)

            return (
                self.name
                + "("
                + ", ".join(
                    [
                        f"{child_type.first} := {child.sql()}"
                        for child, child_type in zip(
                            self.children,
                            result_struct.type_info.child_types,
                            strict=True,
                        )
                    ]
                )
                + ")"
            )
        elif self.is_operator:
            return self.name.join([child.sql() for child in self.children])
        else:
            return f"{self.name}({', '.join([child.sql() for child in self.children])})"


class BoundComparison(Base):
    """
    Represents a comparison expression, which compares two expressions using
    various comparison operators.
    """

    left: Any
    right: Any

    type: Literal[
        "COMPARE_EQUAL",
        "COMPARE_NOTEQUAL",
        "COMPARE_LESSTHAN",
        "COMPARE_GREATERTHAN",
        "COMPARE_LESSTHANOREQUALTO",
        "COMPARE_GREATERTHANOREQUALTO",
        "COMPARE_DISTINCT_FROM",
        "COMPARE_NOT_DISTINCT_FROM",
    ]

    def sql(self) -> str:
        operator = comparison_type_to_sql_operator.get(self.type)
        if operator is None:
            raise ValueError(f"Unsupported comparison type: {self.type}")

        return f"{self.left.sql()} {operator} {self.right.sql()}"


Contents = Annotated[
    Annotated[BoundConstant, Tag("BOUND_CONSTANT")]
    | Annotated[BoundColumnRef, Tag("BOUND_COLUMN_REF")]
    | Annotated[BoundComparison, Tag("BOUND_COMPARISON")]
    | Annotated[BoundCast, Tag("BOUND_CAST")]
    | Annotated[BoundFunction, Tag("BOUND_FUNCTION")]
    | Annotated[BoundOperator, Tag("BOUND_OPERATOR")]
    | Annotated[BoundCase, Tag("BOUND_CASE")]
    | Annotated[BoundBetween, Tag("BOUND_BETWEEN")]
    | Annotated[BoundConjunction, Tag("BOUND_CONJUNCTION")],
    Discriminator(lambda v: v.get("expression_class")),
]


class Expression(BaseWithContext):
    """
    Represents a serialized expression, which can be used to store and transfer
    expressions in a structured format.
    """

    # The parsed contents of the serialized expression.
    contents: Contents


def expression_to_sql(
    *,
    expression: dict[str, Any],
    bound_column_names: list[str],
    bound_column_types: dict[str, str],
) -> str:
    """
    Convert a DuckDB serialized expression back into SQL, with the types of the
    columns tracked.
    """

    context = Context.model_construct(
        bound_column_names=bound_column_names,
        bound_column_types=bound_column_types,
    )

    base = Expression.model_validate({"contents": expression})
    inject_context(base, context)
    return base.contents.sql()


def convert_to_sql(
    source: list[dict[str, Any]], bound_column_names: list[str]
) -> tuple[str, dict[str, str]]:
    bound_column_types: dict[str, str] = {}
    sql = " AND ".join(
        [
            expression_to_sql(
                expression=filter,
                bound_column_names=bound_column_names,
                bound_column_types=bound_column_types,
            )
            for filter in source
        ]
    )
    return sql, bound_column_types
